# Creates game mode config based on type and mode number
def game_mode(game_type: int, mode: int):
    return ('+game_type', str(game_type), '+game_mode', str(mode))


# Casual game mode config (list)
CASUAL = game_mode(game_type=0, mode=0)

# Deathmatch game mode config (list)
DEATHMATCH = game_mode(game_type=1, mode=2)



# r/GlobalOffensive/comments/cjhcpy/game_state_integration_a_very_large_and_indepth


# Game state integration round descriptor
class Round(dict):
    # The phase of the current round. Value is freezetime during the initial
    # freeze time as well as team timeouts, live when the round is live, and
    # over when the round is over and players are waiting for the next round to
    # begin.
    @property
    def phase(self):
        return self.get('phase')

    # The winning team of the round.
    @property
    def win_team(self):
        return self.get('win_team')

    # The current state of the bomb. This section will not appear until the bomb
    # has at least been planted.
    @property
    def bomb(self):
        return self.get('bomb')



"""
Counter-Strike: Global Offensive Game State Integration Server
"""

# Run server in separate thread
import threading
# HTTP server (endpoint for game state integration POST requests)
from flask import Flask, request
# Game state integration root game state descriptor
from .gamestate import GameState


# Counter Strike: Global Offensive Game State Integration Server
#   HTTP POST endpoint
class GSIServer:
    """
    HTTP POST request endpoint server for Counter-Strike: Global Offensive Game
    State Integration requests run in a separate thread.
    """

    # Configures game state integration service
    def __init__(self, path, port):
        """
        Initializes the HTTP server, current game state and thread lock for
        accessing the game state
        """
        # Current game state
        self.state = None
        # Thread lock to synchronize access to the game state
        self.lock = threading.Lock()

        # Server thread
        def server():
            # Setup flask http service
            _server = Flask(__name__)

            # Handle HTTP POST request
            @_server.route(path, methods=['POST'])
            def post():
                # Lock access to the game state
                with self.lock:
                    # Interpret request as json and write to wrapping object
                    self.state = request.get_json()
                # Send response
                return 'OK'

            # Run the flask app
            _server.run(port=port)

        # Create and start server thread
        threading.Thread(target=server, daemon=True).start()

    # Gets next game state integration if available
    def grab(self, reset=False, block=False):
        """
        Grabs the current game state.
        :param reset: Reset the state to None after the access
        :param block: Block while there is None game state
        :return: Returns the current game state as a GameState object
        """
        # Wait for new current state
        while block and not self.state:
            pass
        # Get the current state
        state = self.state
        # Reset current state (if not already being set by server thread)
        if reset and self.lock.acquire(blocking=False):
            # Set state back to None (empty)
            self.state = None
            # Release access to the game state
            self.lock.release()
        # Return the current state
        return GameState(state)

    # Streams game state integration
    def stream(self, **kwargs):
        """
        Streams (yields) game states for ever
        :param kwargs: Keyword arguments passed to the grab method
        :return: Yields the next game state as a GameState object
        """
        # Grab game states for ever
        while True:
            # Next game state
            yield self.grab(**kwargs)



# r/GlobalOffensive/comments/cjhcpy/game_state_integration_a_very_large_and_indepth


# Game state integration player
class Player(dict):
    # Player state
    class State(dict):
        # The HP of the currently spectated player. Value 0-100.
        @property
        def health(self):
            return self.get('health')

        # The Armor of the currently spectated player. Value 0-100.
        @property
        def armor(self):
            return self.get('armor')

        # Boolean value showing whether or not the player is wearing a helmet.
        @property
        def helmet(self):
            return self.get('helmet')

        # A value showing how flashed the player is, from 0 (not flashed),
        # to 255 (fully flashed)
        @property
        def flashed(self):
            return self.get('flashed')

        # How obscured the player's vision is from smoke. Value 0-255.
        @property
        def smoked(self):
            return self.get('smoked')

        # Same as above, but for when a player is on fire. Value 0-255, however
        # only 255 (on fire) and 254-0 (not on fire) appear to be relevant, as a
        # player does not stay on fire after leaving it. Still, this value does
        # steadily decrease back down to 0 after leaving fire, which could be
        # useful in on-screen animations, etc.
        @property
        def burning(self):
            return self.get('burning')

        # Current money of the spectated player with no comma separation.
        # Value 0-16000.
        @property
        def money(self):
            return self.get('money')

        # Amount of kills for the spectated player in the current round.
        # Value 0-5 (unsure if TKs increase this value)
        @property
        def round_kills(self):
            return self.get('round_kills')

        # Amount of kills from a head shot for the spectated player in the
        # current round. Value 0-5.
        @property
        def round_killhs(self):
            return self.get('round_killhs')

        # Total value of the currently spectated player's equipment.
        @property
        def equip_value(self):
            return self.get('equip_value')

    # Player match statistics
    class MatchStats(dict):
        # Number of kills of the currently spectated player.
        @property
        def kills(self):
            return self.get('kills')

        # Number of assists.
        @property
        def assists(self):
            return self.get('assists')

        # Number of deaths.
        @property
        def deaths(self):
            return self.get('deaths')

        # Number of MVPs.
        @property
        def mvps(self):
            return self.get('mvps')

        # Current score.
        @property
        def score(self):
            return self.get('score')

    # Player weapons list
    class Weapons(dict):
        # Single weapon descriptor
        class Weapon(dict):
            # The internal name of the weapon held in the weapon slot.
            @property
            def name(self):
                return self.get('name')

            # An internally used name for the weapon skin, or "default" if no
            # skin.
            @property
            def paintkit(self):
                return self.get('paintkit')

            # Type of weapon. Can be "Pistol", "Knife", "Rifle", "SniperRifle",
            # "Submachine Gun", "C4", possibly others.
            @property
            def type(self):
                return self.get('type')

            # Current amount of ammo in the clip.
            @property
            def ammo_clip(self):
                return self.get('ammo_clip')

            # Maximum amount of ammo the clip for the weapon can hold.
            @property
            def ammo_clip_max(self):
                return self.get('ammo_clip_max')

            # Amount of extra (reserve) ammo available for the weapon.
            @property
            def ammo_reserve(self):
                return self.get('ammo_reserve')

            # Current state of the weapon, "active" if currently being used,
            # "holstered" if not.
            @property
            def state(self):
                return self.get('state')

        # Iterates all weapons
        def __iter__(self):
            return (Player.Weapons.Weapon(x) for (_, x) in self.items())

        # Gets the weapon at specified slot
        def slot(self, index):
            # Query weapon at slot
            weapon = self.get(f'weapon_{index}')
            # If weapon at the slot exists
            if weapon is not None:
                # Wrap in Weapon object
                return Player.Weapons.Weapon(weapon)
            # No weapon at the slot
            return None

        # The players active weapon
        @property
        def active(self):
            # Iterate all players weapons
            for (key, value) in self.items():
                # Test weapon state (only one should be active)
                if value['state'] == 'active':
                    # Return the active weapon
                    return key, Player.Weapons.Weapon(value)
            # No weapon is active
            return None

        # The players grenades
        @property
        def grenades(self):
            # Filter by weapon type
            data = {k: v for (k, v) in self.items() if v['type'] == 'Grenade'}
            # Create new weapons wrapper
            return Player.Weapons(data)

    # Player name
    @property
    def name(self):
        return self.get('name')

    # Player clan name
    @property
    def clan(self):
        return self.get('clan')

    # Player steamid
    @property
    def steamid(self):
        return self.get('steamid')

    # Player observer slot when spectating player
    @property
    def observer_slot(self):
        return self.get('observer_slot')

    # Player team (either 'T' or 'CT', maybe None in menu)
    @property
    def team(self):
        return self.get('team')

    # Current player activity (playing, menu or textinput)
    @property
    def activity(self):
        return self.get('activity')

    # The SteamID of the spectated player in SteamID64 format
    @property
    def spectarget(self):
        return self.get('spectarget')

    # Player position (x,y,z)
    @property
    def position(self):
        # Query position field (holds string, might not exist)
        position = self.get('position')
        # If position field exists
        if position is not None:
            # Split coordinates and convert to float
            return list(map(float, position.split(',')))
        # No player position
        return None

    # Player forward movement/direction
    @property
    def forward(self):
        # Query forward field (holds string, might not exist)
        forward = self.get('forward')
        # If forward field exists
        if forward is not None:
            # Split coordinates and convert to float
            return list(map(float, forward.split(',')))
        # No player forward
        return None

    # Player state (health, armor, flashed, ...)
    @property
    def state(self):
        # Query player state (might not exist)
        state = self.get('state')
        # If state exists, wrap in State object
        if state is not None:
            return Player.State(state)
        # No player state fields
        return None

    # Player match statistics (kills, assists, deaths, ...)
    @property
    def match_stats(self):
        # Query player match_stats (might not exist)
        match_stats = self.get('match_stats')
        # If match_stats exists, wrap in State object
        if match_stats is not None:
            return Player.MatchStats(match_stats)
        # No player match_stats fields
        return None

    # Player weapons
    @property
    def weapons(self):
        # Query player weapons (might not exist)
        weapons = self.get('weapons')
        # If weapons exists, wrap in Weapons object
        if weapons is not None:
            return Player.Weapons(weapons)
        # No player weapons fields
        return None



# Environment variable
from os import environ
# Caching the display properties for increased performance
from functools import cached_property
# Starting a subprocess
from subprocess import Popen
# Context (with statement)
from contextlib import contextmanager
# Numpy arrays
import numpy as np
# Import mss screen capture
import mss
# Import pynput mouse subpackage
from pynput import mouse
# Import pynput keyboard subpackage
from pynput import keyboard


# Virtual display interface
class VirtualDisplay:
    # Registry of known display methods
    _registry = {}

    # Initializes subclasses to add to registry of implemented virtual display
    # setup methods
    def __init_subclass__(cls, method, **kwargs):
        super().__init_subclass__(**kwargs)
        # Add construction method to known display methods
        cls._registry[method] = cls

    # Constructs a class of given method
    def __new__(cls, method, display, *args, **kwargs):
        # Lookup class implementing method in registry
        subclass = cls._registry[method]
        # Construct instance of selected subclass
        obj = object.__new__(subclass)
        # Set empty display number and display process
        obj.display = display
        obj.x = None
        # Return constructed object
        return obj

    # Initializes an virtual display
    def __init__(self, method, display, *args, **kwargs):
        # Store method selector
        self.method = method
        # Set empty display number and display process
        self.display = display
        self.x = None

    # Gets pynput mouse controller for controlling the virtual display
    @cached_property
    def mouse(self):
        # Remember currently active display
        DISPLAY = environ['DISPLAY']
        # Activate the virtual display
        environ['DISPLAY'] = self.display
        # Create mouse controller connected to the virtual display (this reads
        # current environ['DISPLAY']
        controller = mouse.Controller()
        # Reset active display
        environ['DISPLAY'] = DISPLAY
        # Return mouse controller object
        return controller

    # Gets pynput keyboard controller for controlling the virtual display
    @cached_property
    def keyboard(self):
        # Remember currently active display
        DISPLAY = environ['DISPLAY']
        # Activate the virtual display
        environ['DISPLAY'] = self.display
        # Create keyboard controller connected to the virtual display (this
        # reads current environ['DISPLAY']
        controller = keyboard.Controller()
        # Reset active display
        environ['DISPLAY'] = DISPLAY
        # Return keyboard controller object
        return controller

    # Gets mss screen for capturing the virtual display
    @cached_property
    def screen(self):
        # Return mss screen capture object
        return mss.mss(display=self.display)

    # Captures an image of the virtual screen
    def capture(self):
        # Grab current screen output and convert to numpy (compatible with cv2)
        return np.array(self.screen.grab(self.screen.monitors[0]))[:, :, :3]

    # Activates the display
    @contextmanager
    def activate(self):
        # Remember currently active display
        DISPLAY = environ['DISPLAY']
        # Activate the virtual display
        environ['DISPLAY'] = self.display
        # Back to call side (with statement)
        yield
        # Reset active display
        environ['DISPLAY'] = DISPLAY

    # Closes virtual display
    def close(self):
        # If there is a virtual display process
        if self.x is not None:
            # Terminate display process
            self.x.terminate()
            # Reset display process
            self.x = None


# Virtual display class using Xvfb method
class XvfbVirtualDisplay(VirtualDisplay, method='Xvfb'):
    # Initializes the virtual display
    def __init__(self, method, width=640, height=480, depth=24, display=':99'):
        super().__init__(method, display)
        # Store display number
        self.display = display
        # Start virtual display process depending on method
        self.x = Popen(['Xvfb', display, '-screen', '0',
            f'{width}x{height}x{depth}'])


# Virtual display class using Xephyr method
class XephyrVirtualDisplay(VirtualDisplay, method='Xephyr'):
    # Initializes the virtual display
    def __init__(self, method, width=640, height=480, depth=24, display=':99'):
        super().__init__(method, display)
        # Store display number
        self.display = display
        # Start virtual display process depending on method
        self.x = Popen(['Xephyr', display, '-screen',
            f'{width}x{height}x{depth}'])



# r/GlobalOffensive/comments/cjhcpy/game_state_integration_a_very_large_and_indepth


# Game state integration phase_countdowns descriptor
class PhaseCountdowns(dict):
    # The current phase of the round. Values are similar to the "round"
    # component (live, over, freezetime), but also includes "bomb" when the bomb
    # is planted, "defuse" when the bomb is being defused, and "warmup" during
    # the pre-game warmup time.
    @property
    def phase( self ):
        return self.get('phase')
    
    # The time in which the current phase ends. This timer changes depending on
    # the phase. For example, during "live" it will display the time left in the
    # round, but when the bomb is planted it will change to display the time
    # before the bomb explodes. When the round ends and phase goes to "over", it
    # will display the time left before the new round starts.
    @property
    def phase_ends_in( self ):
        return self.get('phase_ends_in')



# Provider gamestate descriptor
from .provider import Provider
# Player game state descriptor
from .player import Player
# Bomb game state descriptor
from .bomb import Bomb
# Round game state descriptor
from .round import Round
# PhaseCountdowns game state descriptor
from .phase_countdowns import PhaseCountdowns
# Map game state descriptor
from .map import Map


# Global game state root structure
class GameState(dict):
    # Initializes game state dict from data handling None
    def __init__(self, data, **kwargs):
        # If no data given
        if data is None:
            # Use empty dict
            data = {}
        # Initialize dict
        super().__init__(data, **kwargs)

    # Gets game state provider data
    @property
    def provider(self):
        # Query provider data (might not exist)
        provider = self.get('provider')
        # If provider field exists
        if provider is not None:
            # Wrap in provider object
            return Provider(provider)
        # No provider
        return None

    # Gets game state player data if present
    @property
    def player(self):
        # Query player data (might not exist)
        player = self.get('player')
        # If player field exists
        if player is not None:
            # Wrap in player object
            return Player(player)
        # No player
        return None

    # Gets game state bomb data if present
    @property
    def bomb(self):
        # Query bomb data (might not exist)
        bomb = self.get('bomb')
        # If bomb field exists
        if bomb is not None:
            # Wrap in bomb object
            return Bomb(bomb)
        # No bomb
        return None

    # Gets game state round data if present
    @property
    def round(self):
        # Query round data (might not exist)
        _round = self.get('round')
        # If round field exists
        if _round is not None:
            # Wrap in round object
            return Round(_round)
        # No round
        return None

    # Gets game state phase_countdowns data if present
    @property
    def phase_countdowns(self):
        # Query phase_countdowns data (might not exist)
        phase_countdowns = self.get('phase_countdowns')
        # If phase_countdowns field exists
        if phase_countdowns is not None:
            # Wrap in phase_countdowns object
            return PhaseCountdowns(phase_countdowns)
        # No phase_countdowns
        return None

    # Gets game state map data if present
    @property
    def map(self):
        # Query map data (might not exist)
        _map = self.get('map')
        # If map field exists
        if _map is not None:
            # Wrap in map object
            return Map(_map)
        # No map
        return None

    # Gets new game state data added this frame
    @property
    def added(self):
        # Query added data (might not exist)
        added = self.get('added')
        # If added field exists
        if added is not None:
            # Wrap in GameState object
            return GameState(added)
        # No added data
        return None

    # Gets previous data (changed this frame)
    @property
    def previously(self):
        # Query previously data (might not exist)
        previously = self.get('previously')
        # If previously field exists
        if previously is not None:
            # Wrap in GameState object
            return GameState(previously)
        # No previously data
        return None



# Pynput library to interact with the keyboard
from pynput import keyboard


# Sets keyboard key state
def set_key( controller, key, state ):
    # If state is True (Pressed)
    if state:
        # Press the key
        controller.press(key)
    else:
        # Otherwise release the key
        controller.release(key)


# Sets Counter Strike: Global Offensive keyboard key state
def control_keyboard( controller, state ):
    # Set move forward state
    set_key(controller, 'W', state['forward'])
    # Set move left state
    set_key(controller, 'A', state['left'])
    # Set move back state
    set_key(controller, 'S', state['back'])
    # Set move right state
    set_key(controller, 'D', state['right'])
    # Set walk state
    set_key(controller, keyboard.Key.shift, state['walk'])
    # Set jump state
    set_key(controller, keyboard.Key.space, state['jump'])
    # Set duck state
    set_key(controller, keyboard.Key.ctrl, state['duck'])
    # Set reload state
    set_key(controller, 'R', state['reload'])
    # Set drop state
    set_key(controller, 'G', state['drop'])
    # Set use state
    set_key(controller, 'E', state['use'])
    # Set switch state
    set_key(controller, 'Q', state['switch'])
    # Set healthshot state
    set_key(controller, 'X', state['healthshot'])
    # Set equip state
    for key in range(10):
        set_key(controller, str(key), state['equip'] == key)


# Resets Counter Strike: Global Offensive keyboard key state
def reset_keyboard( controller ):
    # Set move forward state
    set_key(controller, 'W', False)
    # Set move left state
    set_key(controller, 'A', False)
    # Set move back state
    set_key(controller, 'S', False)
    # Set move right state
    set_key(controller, 'D', False)
    # Set walk state
    set_key(controller, keyboard.Key.shift, False)
    # Set jump state
    set_key(controller, keyboard.Key.space, False)
    # Set duck state
    set_key(controller, keyboard.Key.ctrl, False)
    # Set reload state
    set_key(controller, 'R', False)
    # Set drop state
    set_key(controller, 'G', False)
    # Set use state
    set_key(controller, 'E', False)
    # Set switch state
    set_key(controller, 'Q', False)
    # Set healthshot state
    set_key(controller, 'X', False)
    # Set equip state
    for key in range(10):
        set_key(controller, str(key), False)



# Launch a subprocess
import subprocess
# Import os functions
import os
# Signal number (SIGTERM)
import signal

# Wraps csgo process group (session)
class CSGOSessionWrapper():
    # Constructs wrapper of Popen object
    def __init__(self, popen):
        self.popen = popen

    # Terminates process group
    def terminate(self):
        # Kill process group of popen object with SIGTERM
        os.killpg(os.getpgid(self.popen.pid), signal.SIGTERM)

    # Kills process group
    def kill(self):
        # Kill process group of Popen object with SIGKILL
        os.killpg(os.getpgid(self.popen.pid), signal.SIGKILL)

    # Gets the id of the process group
    def getpgid(self):
        # Get group id of child process
        return os.getpgid(self.popen.pid)


# Launch Counter Strike: Global Offensive in subprocess
def launch_csgo(args=None):
    # Immutable argument
    if args is None:
        args = []
    # Launch Counter Strike process via steam
    process = subprocess.Popen(['steam', '-applaunch', '730', *args],
        start_new_session=True)
    # Return subprocess (e.g. to terminate later) wrapper
    return CSGOSessionWrapper(process)



# OpenAI gym environments
import gym
# Numpy arrays
import numpy as np
# Game weapon specifications
from gym_csgo.specs.weapon import WEAPON_NAMES, WEAPON_TYPES, NUM_WEAPON_SLOTS
# Enumeration observation space
from .enum import Enum


# Constructs a Counter Strike weapon observation space
def make_weapon_obs_space():
    # Observation space base components
    from gym.spaces import Box
    # Return observation space dictionary
    return gym.spaces.Dict({
        # Name of the weapon
        'name': Enum(*WEAPON_NAMES, 'none'),
        # Type of the weapon
        'type': Enum(*WEAPON_TYPES, 'none'),
        # State of weapon
        'state': Enum('active', 'holstered', 'none'),
        # Current amount of ammo in the clip
        'ammo_clip': Box(low=0, high=np.inf, shape=(1,)),
        # Maximum amount of ammo the clip for the weapon can hold
        'ammo_clip_max': Box(low=0, high=np.inf, shape=(1,)),
        # Amount of extra (reserve) ammo available for the weapon
        'ammo_reserve': Box(low=0, high=np.inf, shape=(1,))
    })


# Constructs a Counter Strike weapon observation space
def make_obs_space(pov_shape=(640, 480, 3)):
    # Observation space base components
    from gym.spaces import Box, Discrete
    # Get the weapon observation space
    weapon  = make_weapon_obs_space()
    # Return observation space dictionary
    return gym.spaces.Dict({
        # Observation image
        'pov': Box(low=0, high=255, shape=pov_shape, dtype=np.uint8),
        # Player state
        'health': Box(low=0, high=np.inf, shape=(1,)),
        'armor': Box(low=0, high=np.inf, shape=(1,)),
        'helmet': Discrete(2),
        'flashed': Box(low=0, high=255, shape=(1,), dtype=np.uint8),
        'smoked': Box(low=0, high=255, shape=(1,), dtype=np.uint8),
        'burning': Box(low=0, high=255, shape=(1,), dtype=np.uint8),
        'money': Box(low=0, high=np.inf, shape=(1,)),
        'value': Box(low=0, high=np.inf, shape=(1,)),
        # Player stats
        'kills': Box(low=0, high=np.inf, shape=(1,)),
        'assists': Box(low=0, high=np.inf, shape=(1,)),
        'deaths': Box(low=0, high=np.inf, shape=(1,)),
        'mvps': Box(low=0, high=np.inf, shape=(1,)),
        'score': Box(low=0, high=np.inf, shape=(1,)),
        # The player's team
        'team': Enum('CT', 'T', 'none'),
        # The player's active weapon
        'weapon_active': weapon,
        # All the player's weapons
        **{f'weapon_{i}': weapon for i in range(NUM_WEAPON_SLOTS)},
    })



# Player game state integration
from gym_csgo.gsi.player import Player


# List of known CSGO weapon names
WEAPON_NAMES = ['weapon_' + name for name in [
    'c4', 'knife', 'knife_t', 'taser', 'shield', 'bumpmine', 'breachcharge',
    'decoy', 'flashbang', 'healthshot', 'hegrenade', 'incgrenade', 'molotov',
    'smokegrenade', 'tagrenade', 'm249', 'mag7', 'negev', 'nova', 'sawedoff',
    'xm1014', 'cz75a', 'deagle', 'elite', 'fiveseven', 'glock', 'hkp2000',
    'p250', 'revolver', 'tec9', 'usp_silencer', 'ak47', 'aug', 'awp', 'famas',
    'g3sg1', 'galilar', 'm4a1', 'm4a1_silencer', 'scar20', 'sg556', 'ssg08',
    'bizon', 'mac10', 'mp5sd', 'mp7', 'mp9', 'p90', 'ump45'
]]

# List of known CSGO weapon types
WEAPON_TYPES = [
    'Pistol', 'Knife', 'Rifle', 'SniperRifle', 'Submachine Gun', 'C4', 'Grenade'
]

# Weapon slots to consider
NUM_WEAPON_SLOTS = 10

# Observes a weapon converts from GSI object to dict weapon observation
def observe_weapon(weapon: Player.Weapons.Weapon) -> dict:
    # No weapon supplied
    if weapon is None:
        return {}

    # Extracts value or return 'none' enum
    def value_or_none(value):
        return value if value is not None else 'none'

    # Construct dictionary
    return {
        # Discrete Enum values
        'name': value_or_none(weapon.name),
        'type': value_or_none(weapon.type),
        'state': value_or_none(weapon.state),
        # Continuous Box values
        'ammo_clip': value_or_none(weapon.ammo_clip),
        'ammo_clip_max': value_or_none(weapon.ammo_clip_max),
        'ammo_reserve': value_or_none(weapon.ammo_reserve)
    }



# r/GlobalOffensive/comments/cjhcpy/game_state_integration_a_very_large_and_indepth


# Game state integration map descriptor
class Map(dict):
    # Map team descriptor
    class Team(dict):
        # The team's current score.
        @property
        def score(self):
            return self.get('score')

        # How many rounds the team has lost in a row.
        @property
        def consecutive_round_losses(self):
            return self.get('consecutive_round_losses')

        # The number of remaining timeouts available for the team.
        @property
        def timeouts_remaining(self):
            return self.get('timeouts_remaining')

        # How many games a team has won in a Best of X series. Only used for
        # tournaments.
        @property
        def matches_won_this_series(self):
            return self.get('matches_won_this_series')

    # Map round_wins descriptor
    class RoundWins(dict):
        # Iterates all rounds
        def __iter__(self):
            return ((_round, win) for (_round, win) in self.items())

        # Get round winning condition for round
        def round(self, index):
            return self.get(f'{index}')

    # The current game mode being played. Values can be casual, competitive,
    # deathmatch, gungameprogressive, scrimcomp2v2, possibly others.
    @property
    def mode(self):
        return self.get('mode')

    # The name of the current map.
    @property
    def name(self):
        return self.get('name')

    # The current phase of the map. Values can be warmup during the initial
    # warmup phase, live during a live game, intermission during halftime, and
    # gameover at the end of the game.
    @property
    def phase(self):
        return self.get('phase')

    # The current round number.
    @property
    def round(self):
        return self.get('round')

    # Team data
    def team(self, key):
        # Query team data (might not exist)
        team = self.get(f'team_{key.lower()}')
        # If team exists, wrap in State object
        if team is not None:
            return Map.Team(team)
        # No team fields
        return None

    # Terrorist team data
    @property
    def team_t(self):
        # Query team data (might not exist)
        team_t = self.get('team_t')
        # If team exists, wrap in State object
        if team_t is not None:
            return Map.Team(team_t)
        # No team fields
        return None

    # Counter Terrorist team data
    @property
    def team_ct(self):
        # Query team data (might not exist)
        team_t = self.get('team_ct')
        # If team exists, wrap in State object
        if team_t is not None:
            return Map.Team(team_t)
        # No team fields
        return None

    # How many matches a team has to win before winning the series.
    @property
    def num_matches_to_win_series(self):
        return self.get('num_matches_to_win_series')

    # Current number of people spectating the game.
    @property
    def current_spectators(self):
        return self.get('current_spectators')

    # How many souvenir cases were dropped this game (Probable, unconfirmed).
    @property
    def souvenirs_total(self):
        return self.get('souvenirs_total')

    # Round wins on this map
    @property
    def round_wins(self):
        # Query round_wins data (might not exist)
        round_wins = self.get('round_wins')
        # If round_wins exists, wrap in State object
        if round_wins is not None:
            return Map.RoundWins(round_wins)
        # No round_wins fields
        return None



# r/GlobalOffensive/comments/cjhcpy/game_state_integration_a_very_large_and_indepth


# Game state integration bomb descriptor
class Bomb(dict):
    # The current state of the bomb. Values are "carried", "dropped",
    # "planting", "planted", "defusing", and "defused"
    @property
    def state(self):
        return self.get('state')

    # The current position of the bomb on the map, in x, y, z coordinates.
    @property
    def position(self):
        # Query position field (holds string, might not exist)
        position = self.get('position')
        # If position field exists
        if position is not None:
            # Split coordinates and convert to float
            return list(map(float, position.split(',')))
        # No bomb position
        return None

    # The bomb's countdown timer. Used as regular bomb timer when
    # state = planted, time until bomb plant when state = planting, and time
    # until defuse when state = defusing.
    @property
    def countdown(self):
        return self.get('countdown')

    # The SteamID of the player interacting with the bomb, in SteamID64 format.
    # The interacting player can be the player carrying the bomb
    # (state = carried), planting the bomb (state = planting), or defusing the
    # bomb (state = defusing).
    @property
    def player(self):
        return self.get('player')



# Gym environments, csgo to register envs
import gym, gym_csgo
# Command line argument parsing
from .args import make_argparse


# Main function entry point (do not want to have stuff in global scope)
def main(env_id, mapname, **kwargs):
    # Open new environment context (automatically closes env at end of scope)
    with gym.make(env_id, mapname=mapname, **kwargs) as env:
        # Reset the environment
        env.reset()
        # Env is not done yet
        done = False
        # Until the environment is done
        while not done:
            # Get random action from environment
            action = env.action_space.noop()
            # Execute the random action and collect observation
            obs, rew, done, info = env.step(action)
            # Show current observation (image)
            env.render()


# Script entry point
if __name__ == '__main__':
    # Construct environment argument parser
    parser = make_argparse()
    # Parse supplied arguments using the parser
    args = parser.parse_args()
    # Call main function and forward the arguments
    main(
        # Environment / Map / Game Setup
        args.env_id, args.mapname, args=args.args,
        # Display Setup
        width=args.width, height=args.height, display=args.display,
        display_method=args.display_method,
        # Game State Integration Setup
        gsi_path=args.gsi_path, gsi_port=args.gsi_port,
    )



# Game state integration provider
class Provider(dict):
    # Provider name
    @property
    def name(self):
        return self.get('name')

    # Provider appid
    @property
    def appid(self):
        return self.get('appid')

    # Provider version
    @property
    def version(self):
        return self.get('version')

    # Provider steamid
    @property
    def steamid(self):
        return self.get('steamid')

    # Provider timestamp (unix time)
    @property
    def timestamp(self):
        return self.get('timestamp')



# Gym environment registration
from gym.envs.registration import register

# Register the csgo environment
register(id='csgo-v0', entry_point='gym_csgo.envs:CSGOEnv')

# Register the csgo casual mode environment
register(id='csgo_casual-v0', entry_point='gym_csgo.envs:CSGOCasualEnv')

# Register the csgo deathmatch mode environment
register(id='csgo_dm-v0', entry_point='gym_csgo.envs:CSGODeathmatchEnv')



# Command line argument parser
import argparse


# Constructs argument parser parsing all arguments required for CSGOEnv
def make_argparse():
    # Setup command line argument parser
    parser = argparse.ArgumentParser()
    # Positional arguments (all but env_id are optional)
    parser.add_argument(
        'env_id', type=str, choices=('csgo_dm-v0', 'csgo_casual-v0')
    )
    parser.add_argument(
        'mapname', type=str, nargs='?', default='de_dust2'
    )
    parser.add_argument(
        'args', type=str, nargs='*'
    )
    # Display configuration (optional with defaults)
    parser.add_argument(
        '--width', type=int, default=640
    )
    parser.add_argument(
        '--height', type=int, default=480
    )
    parser.add_argument(
        '--display', type=str, default=':99'
    )
    parser.add_argument(
        '--display-method', type=str, choices=('Xvfb','Xephyr'), default='Xvfb'
    )
    # Game State Integration configuration (optional with defaults)
    parser.add_argument(
        '--gsi-path', type=str, default='/csgo-gsi'
    )
    parser.add_argument(
        '--gsi-port', type=int, default=1234
    )
    # Return constructed argument parser
    return parser



# Pynput library to interact with the mouse
from pynput import mouse

# Sets mouse button state
def set_button( controller, button, state ):
    # If state is True (Pressed)
    if state:
        # Press the mouse button
        controller.press(button)
    else:
        # Otherwise release the mouse button
        controller.release(button)

# Sets Counter Strike: Global Offensive mouse state
def control_mouse( controller, state ):
    # Set special state
    set_button(controller, mouse.Button.right, state['special'])
    # Set fire state
    set_button(controller, mouse.Button.left, state['fire'])
    # Move the mouse pointer (moves the game camera)
    controller.move(*state['camera'])

# Resets Counter Strike: Global Offensive mouse state
def reset_mouse( controller ):
    # Set special state
    set_button(controller, mouse.Button.right, False)
    # Set fire state
    set_button(controller, mouse.Button.left, False)



from setuptools import setup, find_packages

# Read long description from README
with open("README.md", "r", encoding="utf-8") as file:
    long_description = file.read()

# Setup the python package with dependencies
setup(
    # Package name and version and contained packages
    name='gym-csgo', version='0.0.2', packages=find_packages(),
    # Package metadata
    description="Counter-Strike: Global Offensive environment for OpenAI Gym "
                "on Linux",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/iksnagreb/gym-csgo",
    author="Christoph Berganski",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: POSIX :: Linux"
    ],
    # Package requirements
    install_requires=[
        'gym', 'mss', 'numpy', 'opencv-python', 'flask', 'pynput', 'pygame'
    ],
    # Add non-code files to package
    include_package_data=True,
)



# OpenAI gym environments
import gym


# Constructs a Counter Strike action space
def make_act_space():
    # Return action space dictionary
    return gym.spaces.Dict({
        'forward': gym.spaces.Discrete(2),
        'back': gym.spaces.Discrete(2),
        'left': gym.spaces.Discrete(2),
        'right': gym.spaces.Discrete(2),
        'walk': gym.spaces.Discrete(2),
        'jump': gym.spaces.Discrete(2),
        'duck': gym.spaces.Discrete(2),
        'fire': gym.spaces.Discrete(2),
        'special': gym.spaces.Discrete(2),
        'reload': gym.spaces.Discrete(2),
        'drop': gym.spaces.Discrete(2),
        'use': gym.spaces.Discrete(2),
        'switch': gym.spaces.Discrete(2),
        'healthshot': gym.spaces.Discrete(2),
        'equip': gym.spaces.Discrete(10),
        'camera': gym.spaces.Box(low=-540.0, high=540.0, shape=(2,))
    })


# Creates a noop (do nothing) action
def make_noop():
    # Return dictionary with all action set to 0 (False)
    return {
        # Do not move
        'forward': 0, 'back': 0, 'left': 0, 'right': 0,
        # No sneaking, hiding, ...
        'walk': 0, 'jump': 0, 'duck': 0,
        # No interacting
        'fire': 0, 'special': 0, 'reload': 0, 'drop': 0, 'use': 0,
        # No weapon change
        'switch': 0, 'healthshot': 0, 'equip': None,
        # Do not look around
        'camera': [0, 0]
    }



# Interactive game window
import pygame
# Image processing
import cv2
# Gym environments, csgo to register envs
import gym, gym_csgo
# Command line argument parsing
from .args import make_argparse


# Key binding (keyboard layout, pygame - CSGO action mapping)
def map_keys_to_action(keys, action: dict):
    # Mapping dictionary:
    #   Lookup pygame key for each available action
    mappings = {
        'forward': pygame.K_w,
        'left': pygame.K_a,
        'back': pygame.K_s,
        'right': pygame.K_d,
        'jump': pygame.K_SPACE,
        'walk': pygame.K_LSHIFT,
        'duck': pygame.K_LCTRL,
        'reload': pygame.K_r,
        'drop': pygame.K_g,
        'use': pygame.K_e,
        'switch': pygame.K_q,
        'healthshot': pygame.K_x
    }
    # Iterate all actions and test whether the mapped key is set
    for key in mappings:
        action[key] = keys[mappings[key]]
    # Equip weapon selections
    selections = [
        pygame.K_0, pygame.K_1, pygame.K_2, pygame.K_3, pygame.K_4, pygame.K_5,
        pygame.K_6, pygame.K_7, pygame.K_8, pygame.K_9
    ]
    # Iterate all selections and set the equip action
    for (index,key) in enumerate(selections):
        # If the key is pressed
        if keys[key]:
            # Update equipment selection with index
            action["equip"] = index
    # Return updated action
    return action


# Maps mouse state to action
def map_mouse_to_action(mouse, action, scale=50, limit=480):
    # Get mouse movement since last step
    (x, y) = mouse.get_rel()
    # Set camera actions
    action["camera"][0] = min(max(scale * x, -limit), limit)
    action["camera"][1] = min(max(scale * y, -limit), limit)
    # Attack with mouse left click
    action["fire"] = mouse.get_pressed()[0]
    # Special weapon action with mouse right click
    action["special"] = mouse.get_pressed()[2]
    # Return modified action
    return action


# Main function entry point (do not want to have stuff in global scope)
def main(env_id, mapname, **kwargs):
    # Initialize pygame environment
    pygame.init()
    # Setup pygame display for input and output of the environment
    display = pygame.display.set_mode((kwargs['width'], kwargs['height']))
    # Do not grab the input (mouse constrained to the pygame display)
    pygame.event.set_grab(False)
    # Show no mouse cursor
    pygame.mouse.set_visible(False)
    # Open new environment context (automatically closes env at end of scope)
    with gym.make(env_id, mapname=mapname, **kwargs) as env:
        # Reset the environment
        env.reset()
        # Env is not done yet
        done = False
        # Until the environment is done
        while not done:
            # Query input events
            pygame.event.pump()
            # Get keyboard key states
            keys = pygame.key.get_pressed()
            # Toggle mouse grab if ESC pressed
            if keys[pygame.K_ESCAPE]:
                pygame.event.set_grab(not pygame.event.get_grab())
            # Initialize using no-operation
            action = env.action_space.noop()
            # Use keyboard input to set/unset actions
            action = map_keys_to_action(keys, action)
            # Use mouse input to update actions
            action = map_mouse_to_action(pygame.mouse, action)
            # Execute the prepared action and collect observation
            obs, rew, done, _ = env.step(action)
            # Clear the display
            display.fill((255, 255, 255))
            # Convert pov observation to use with pygame
            pov = cv2.cvtColor(obs['pov'], cv2.COLOR_BGR2RGB).swapaxes(0, 1)
            # Display pov observation on pygame display
            display.blit(pygame.pixelcopy.make_surface(pov), (0, 0))
            # Flip the display (somehow the image is rotated/flipped...)
            pygame.display.flip()
            # Show display
            pygame.display.update()


# Script entry point
if __name__ == '__main__':
    # Construct environment argument parser
    parser = make_argparse()
    # Parse supplied arguments using the parser
    args = parser.parse_args()
    # Call main function and forward the arguments
    main(
        # Environment / Map / Game Setup
        args.env_id, args.mapname, args=args.args,
        # Display Setup
        width=args.width, height=args.height, display=args.display,
        display_method=args.display_method,
        # Game State Integration Setup
        gsi_path=args.gsi_path, gsi_port=args.gsi_port,
    )



# OpenAI gym environments
import gym
# Numpy arrays
import numpy as np


# Enum (from string) observation/action space
class Enum(gym.spaces.Discrete):
    # Constructs an Enum space from selection of values
    def __init__(self, *values: str, seed=None):
        # Construct Discrete space super class of number of values
        super().__init__(len(values), seed=seed)
        # Store names of enum values
        self.values = np.array(sorted(values))

    # Samples from enum values
    def sample(self):
        # Use Discrete baseclass sample to select from enum values
        return self.values[super().sample()]

    # Checks whether Enum space contains a value
    def contains(self, x):
        # Test whether item in enum values (list)
        return x in self.values

    # Produces string representation of the enum space
    def __repr__(self):
        return "Enum(" + ','.join(self.values) + ")"

    # Compares to other space (Enum)
    def __eq__(self, other):
        return isinstance(other, Enum) and self.values == other.values



# Sleeping
from time import sleep
# OpenAI gym environments
import gym
# Show image using OpenCV
import cv2
# Virtual display server
from gym_csgo.glue.display import VirtualDisplay
# Mouse controller (decodes action dict)
from gym_csgo.glue.mouse import control_mouse, reset_mouse
# Keyboard controller (decodes action dict)
from gym_csgo.glue.keyboard import control_keyboard, reset_keyboard
# CSGO launcher
from gym_csgo.launch import launch_csgo
# Game state integration server (HTTP endpoint)
from gym_csgo.gsi import GSIServer
# Game startup arguments
from gym_csgo.specs.args import ESSENTIAL_ARGS
# Game weapon specifications
from gym_csgo.specs.weapon import NUM_WEAPON_SLOTS, observe_weapon
# Action space constructor
from gym_csgo.spaces.act import make_act_space, make_noop
# Observation space constructor
from gym_csgo.spaces.obs import make_obs_space


# Counter Strike: Global Offensive environment class
class CSGOEnv(gym.core.Env):
    # Initializes Counter-Strike environment
    def __init__(self, width=640, height=480, display_method='Xephyr',
        display=':99', gsi_path='/csgo-gsi', gsi_port=1234, args=None):
        # Immutable args parameter
        if args is None:
            args = []
        # Startup virtual display server
        self.display = VirtualDisplay(display_method, width=width,
            height=height, depth=24, display=display)
        # Setup the game state integration
        self.gsi = GSIServer(path=gsi_path, port=gsi_port)
        # Setup csgo display arguments
        display_args = [
            '-width', f'{width}', '-height', f'{height}', '-fullscreen'
        ]
        # Prepend default and append user supplied arguments
        self.csgo_args = [*ESSENTIAL_ARGS, *display_args, *args]
        # Launch csgo client session
        self.csgo_session = None
        # Collect sample screenshot
        pov = self.display.capture()
        # Setup observation space with pov shape from first real observation
        self.observation_space = make_obs_space(pov_shape=pov.shape)
        # Setup action space
        self.action_space = make_act_space()
        # Add noop action constructor
        self.action_space.noop = make_noop
        # Last observation
        self.last_obs = None
        # Current observation
        self.current_obs = None
        # Current game state
        self.current_state = None

    # Resets the environment
    def reset(self):
        # Terminate csgo client session
        if self.csgo_session:
            self.csgo_session.terminate()
        # Reset input states
        reset_keyboard(self.display.keyboard)
        reset_mouse(self.display.mouse)
        # Last observation
        self.last_obs = None
        # Current observation
        self.current_obs = None
        # Current game state
        self.current_state = None
        # Activate the virtual display
        with self.display.activate():
            # Launch new csgo client session
            self.csgo_session = launch_csgo(self.csgo_args)
        # Get first observation
        return self._get_obs()

    # Executes one action step in the environment
    def step(self, action, wait=None):
        # Send action to environment
        self._set_act(action)
        # If specified to wait for some time
        if wait is not None:
            # Sleep for requested time
            sleep(wait)
        # Collect new state observation
        obs = self._get_obs()
        # Return new (state,reward,done,info) tuple
        return obs, self._get_rew(), self._is_done(), self._get_info()

    # Closes the environment
    def close(self):
        # Terminate csgo client session
        if self.csgo_session:
            self.csgo_session.terminate()
        # Reset input states
        reset_keyboard(self.display.keyboard)
        reset_mouse(self.display.mouse)
        # Close virtual display
        self.display.close()

    # Renders a view of the environment
    def render(self, mode='human'):
        # Show image of current pov observation
        cv2.imshow(f'{self.__class__.__name__}', self.current_obs['pov'])
        # Do not block
        cv2.waitKey(1)

    # Collects new observation from the environment
    def _get_obs(self):
        # Wait for the environment to get ready
        while not self._is_ready() and not self._is_done():
            self.current_state = self.gsi.grab()
        # Start with observation dict only containing image observation
        obs = {'pov': self.display.capture()}
        # Get current game state
        state = self.gsi.grab()
        # Get player data
        player = state.player
        # Extract player state data from state if present
        if player is not None and player.state is not None:
            # Add player health
            obs['health'] = player.state.health
            # Add player armor
            obs['armor'] = player.state.armor
            # Add player having helmet
            obs['helmet'] = player.state.helmet
            # Add current grenade effects on player
            obs['flashed'] = player.state.flashed
            obs['smoked'] = player.state.smoked
            obs['burning'] = player.state.burning
            # Add player money and equipment value
            obs['money'] = player.state.money
            obs['value'] = player.state.equip_value
        # Extract player stats from state if present
        if player is not None and player.match_stats is not None:
            # Number of kills by player
            obs['kills'] = player.match_stats.kills
            # Number of assists by player
            obs['assists'] = player.match_stats.assists
            # Number of deaths of player
            obs['deaths'] = player.match_stats.deaths
            # Number of MVPs
            obs['mvps'] = player.match_stats.mvps
            # Total score of player
            obs['score'] = player.match_stats.score
        # Extract player team
        obs['team'] = player.team if player.team is not None else 'none'
        # Extract player weapons
        if player is not None and player.weapons is not None:
            # Currently equipped weapon
            if player.weapons.active is not None:
                obs['weapon_active'] = observe_weapon(player.weapons.active[1])
            # Extract weapon slots
            for index in range(NUM_WEAPON_SLOTS):
                obs[f'weapon_{index}'] = observe_weapon(
                    player.weapons.slot(index)
                )
        # Remember last observation
        self.last_obs = self.current_obs
        # Set new current observation
        self.current_obs = obs
        # Set new current game state
        self.current_state = state
        # Return collected observation
        return self.current_obs

    # Sets actions (sends to environment)
    def _set_act(self, action):
        # Send actions only if env is ready
        if self._is_ready():
            # Send actions to mouse controller
            control_mouse(self.display.mouse, action)
            # Send action to keyboard controller
            control_keyboard(self.display.keyboard, action)

    # Gets reward for last step
    def _get_rew(self):
        # Difference of scores
        #   TODO: Probably not a good reward function
        return self.current_obs['score'] - self.last_obs['score']

    # Gets info useful for debugging
    def _get_info(self):
        # Get current game state
        state = self.current_state
        # Get the id of the csgo session process group
        pgid = self.csgo_session.getpgid() if self.csgo_session else None
        # Get the display number
        display = self.display.display if self.display else None
        # Return info dictionary
        return {
            'gsi': state, 'csgo_session_pgid': pgid, 'display': display
        }

    # Tests whether the environment is ready for interactions
    def _is_ready(self):
        # Get current game state
        state = self.current_state
        # For being ready, the map needs to have a phase and players activity
        # needs to be 'playing'
        return state is not None and state.map is not None \
            and state.map.phase == 'live' and state.player.activity == 'playing'

    # Tests whether the environment is done
    def _is_done(self):
        # Get current game state
        state = self.current_state
        # Environment is done, if map phase is 'gameover'
        return state is not None and state.map is not None \
            and state.map.phase == 'gameover'


# Counter Strike: Global Offensive casual game mode environment class
class CSGOCasualEnv(CSGOEnv):
    # Constructs a casual game mode environment
    def __init__(self, mapname='de_dust2', args=None, **kwargs):
        # If no further args supplied, use empty list
        if args is None:
            args = []
        # Casual game mode config
        from gym_csgo.specs.game_modes import CASUAL
        # Construct environment using base class constructor
        super().__init__(args=['+map', mapname, *CASUAL, *args], **kwargs)


# Counter Strike: Global Offensive deathmatch game mode environment class
class CSGODeathmatchEnv(CSGOEnv):
    # Constructs a deathmatch game mode environment
    def __init__(self, mapname='de_dust2', args=None, **kwargs):
        # If no further args supplied, use empty list
        if args is None:
            args = []
        # Casual game mode config
        from gym_csgo.specs.game_modes import DEATHMATCH
        # Construct environment using base class constructor
        super().__init__(args=['+map', mapname, *DEATHMATCH, *args], **kwargs)
